
#include <stdio.h>
#include <stdlib.h>
#include "cell.h"
#include "rif.h"
#include "csi.h"

extern void	Error(/* char *msg */);

#define MA(noe,val)	malloc((size_t)(noe)*sizeof(val))

static Cell	**MakeCell(Nx,Ny,nmfc)
int		Nx,Ny;
char		*nmfc;
{
	Cell	**cell;
	int	y;

	if ((cell=(Cell **)MA(Ny,*cell))==NULL) Error(nmfc);

	for (y=0; y<Ny; y++)
	    if ((cell[y]=(Cell *)MA(Nx,**cell))==NULL) Error(nmfc);

	return cell;
}

#define BITS	(sizeof(Cell)*8)

#define BINARIZE(pv)	\
	if (pv>=lower && pv<=upper) { \
	    ptr[y][x]|=bit; if (x<*X1) *X1=x; if (x>*X2) *X2=x; \
			    if (y<*Y1) *Y1=y; if (y>*Y2) *Y2=y; \
			    if (z<*Z1) *Z1=z; if (z>*Z2) *Z2=z; \
	} \
	else \
	    ptr[y][x]&=off

static int	ReadBW(Nx,Ny,Nz,path,lower,upper,BW,X1,Y1,Z1,X2,Y2,Z2)
int		Nx,Ny,Nz,lower,upper,*X1,*Y1,*Z1,*X2,*Y2,*Z2;
char		**path;
Cell		***BW;
{
	int	z,nx,ny,y,x,pv;
	Cell	**cell,**ptr,bit,off;

	*X1=Nx; *Y1=Ny; *Z1=Nz; *X2=(*Y2)=(*Z2)=(-1);
	for (z=0; z<Nz; z++) {
	    ReadImageFile(path[z],&nx,&ny,NULL,&cell,NULL);

	    ptr=BW[z/BITS]; bit=(Cell)(1<<(z%BITS)); off=(Cell)(~bit);
	    for (y=0; y<Ny; y++) {
		for (x=0; x<Nx; x++) {
		    pv=(x<nx && y<ny)?cell[y][x]:0; BINARIZE(pv);
		}
		if (y<ny) free(cell[y]);
	    }
	    free(cell);
	}
	return *X1<=*X2 && *Y1<=*Y2 && *Z1<=*Z2;
}

static int	GS2BW(Nx,Ny,Nz,GS,lower,upper,BW,X1,Y1,Z1,X2,Y2,Z2)
int		Nx,Ny,Nz,lower,upper,*X1,*Y1,*Z1,*X2,*Y2,*Z2;
Cell		***GS,***BW;
{
	int	z,y,x;
	Cell	**ptr,bit,off;

	*X1=Nx; *Y1=Ny; *Z1=Nz; *X2=(*Y2)=(*Z2)=(-1);
	for (z=0; z<Nz; z++) {
	    ptr=BW[z/BITS]; bit=(Cell)(1<<(z%BITS)); off=(Cell)(~bit);
	    for (y=0; y<Ny; y++)
	    for (x=0; x<Nx; x++) BINARIZE(GS[z][y][x]);
	}
	return *X1<=*X2 && *Y1<=*Y2 && *Z1<=*Z2;
}

typedef unsigned int	Card;

static Card	Search(table,label)
Card		*table,label;
{
	while (table[label]) label=table[label];

	return label;
}

static void	CountCores(X1,Y1,Z1,X2,Y2,Z2,BW,last,title)
int		X1,Y1,Z1,X2,Y2,Z2;
Cell		***BW;
Card		**last;
char		*title;
{
	Card	layer,labels,*table,pixels,label,cores;
	int	x1,y1,z1,x2,y2,z2,x,y,z;
	Cell	**ptr,bit,**ptr2,bit2,**ptr1,bit1,off;

	for (layer=0;; layer++) {
	    x1=X1; y1=Y1; z1=Z1; x2=X2; y2=Y2; z2=Z2;
	    x=x1-1; for (y=y1; y<=y2; y++) last[y][x]=0;
	    y=y1-1; for (x=x1; x<=x2; x++) last[y][x]=0;

	    labels=0;
	    for (y=y1; y<=y2; y++)
	    for (x=x1; x<=x2; x++) last[y][x]=0;
	    for (z=z1; z<=z2; z++) {
		ptr=BW[z/BITS]; bit=(Cell)(1<<(z%BITS));
		for (y=y1; y<=y2; y++)
		for (x=x1; x<=x2; x++)
		    if (ptr[y][x]&bit) {
			if (!last[y][x] && !last[y-1][x] && !last[y][x-1])
			    ++labels;

			last[y][x]=1;
		    }
		    else
			last[y][x]=0;
	    }
	    if ((table=(Card *)MA(labels,*table))==NULL)
		Error("no memory for 'table'.");

	    --table;

	    pixels=0; label=1;
	    for (y=y1; y<=y2; y++)
	    for (x=x1; x<=x2; x++) last[y][x]=0;
	    for (z=z1; z<=z2; z++) {
		ptr=BW[z/BITS]; bit=(Cell)(1<<(z%BITS));
		for (y=y1; y<=y2; y++)
		for (x=x1; x<=x2; x++)
		    if (ptr[y][x]&bit) {
			++pixels;

			if (last[y][x]) {
			    last[y][x]=Search(table,last[y][x]);
if (last[y-1][x] && (last[y-1][x]=Search(table,last[y-1][x]))!=last[y][x])
				last[y-1][x]=table[last[y-1][x]]=last[y][x];
if (last[y][x-1] && (last[y][x-1]=Search(table,last[y][x-1]))!=last[y][x])
				last[y][x-1]=table[last[y][x-1]]=last[y][x];
			}
			else if (last[y-1][x]) {
			    last[y][x]=Search(table,last[y-1][x]);
			    if (last[y][x-1] && last[y][x-1]!=last[y][x])
				last[y][x-1]=table[last[y][x-1]]=last[y][x];
			}
			else if (last[y][x-1])
			    last[y][x]=last[y][x-1];
			else
			    table[last[y][x]=label++]=0;
		    }
		    else
			last[y][x]=0;
	    }
	    cores=0; while (--label>=1) if (!table[label]) ++cores;

	    free(table+1);

	    printf("%s%lu\t%lu\t%lu\n",title,(unsigned long)layer,
					     (unsigned long)pixels,
					     (unsigned long)cores);

	    X1=x2+1; Y1=y2+1; Z1=z2+1; X2=x1-1; Y2=y1-1; Z2=z1-1;
	    ptr2=BW[z1/BITS]; bit2=(Cell)(1<<(z1%BITS));
	    for (y=y1; y<=y2; y++)
	    for (x=x1; x<=x2; x++) last[y][x]=0;
	    for (z=z1; z<=z2; z++) {
		ptr1=ptr2; bit1=bit2; off=(Cell)(~bit1);
		ptr2=BW[(z+1)/BITS]; bit2=(Cell)(1<<((z+1)%BITS));
		for (y=y1; y<=y2; y++)
		for (x=x1; x<=x2; x++)
		    if (ptr1[y][x]&bit1) {
			if (!last[y][x] ||
			    !last[y-1][x] ||
			    !last[y][x-1] ||
			    z==z2 || !(ptr2[y][x]&bit2) ||
			    y==y2 || !(ptr1[y+1][x]&bit1) ||
			    x==x2 || !(ptr1[y][x+1]&bit1))
			    ptr1[y][x]&=off;
			else {
			    if (x<X1) X1=x; if (x>X2) X2=x;
			    if (y<Y1) Y1=y; if (y>Y2) Y2=y;
			    if (z<Z1) Z1=z; if (z>Z2) Z2=z;
			}
			last[y][x]=1;
		    }
		    else
			last[y][x]=0;
	    }
	    if (X1>X2 || Y1>Y2 || Z1>Z2) break;
	}
}

#define TITLE_LEN	256

int	main(argc,argv)
int	argc;
char	**argv;
{
	char	**path,title[TITLE_LEN],*nmfl="no memory for 'last'.",
					*nmfbw="no memory for 'BW'.",
					*nmfgs="no memory for 'GS'.";
	int	Nx,Ny,Nz,y,z,X1,Y1,Z1,X2,Y2,Z2,nx,ny,x,lower,upper;
	Card	**last;
	Cell	***BW,***GS,**cell;

	if (argc!=3 && argc!=5)
	    Error("usage : cc_3d directory nameFile {lower upper}");

	path=CheckSliceImages(argv[1],argv[2],&Nx,&Ny,&Nz,NULL);

	if ((last=(Card **)MA(Ny+1,*last))==NULL) Error(nmfl);

	++last;
	for (y=(-1); y<Ny; y++) {
	    if ((last[y]=(Card *)MA(Nx+1,**last))==NULL) Error(nmfl);

	    ++(last[y]);
	}

	if ((BW=(Cell ***)MA(Nz/BITS+1,*BW))==NULL) Error(nmfbw);

	for (z=0; z<Nz; z+=BITS) BW[z/BITS]=MakeCell(Nx,Ny,nmfbw);

	if (argc==5) {
	    if (!ReadBW(Nx,Ny,Nz,path,atoi(argv[3]),
				      atoi(argv[4]),BW,&X1,&Y1,&Z1,&X2,&Y2,&Z2))
		Error("no object pixel.");

	    CountCores(X1,Y1,Z1,X2,Y2,Z2,BW,last,"");
	}
	else {
	    if ((GS=(Cell ***)MA(Nz,*GS))==NULL) Error(nmfgs);

	    for (z=0; z<Nz; z++) {
		ReadImageFile(path[z],&nx,&ny,NULL,&cell,NULL);

		if (nx==Nx && ny==Ny)
		    GS[z]=cell;
		else {
		    GS[z]=MakeCell(Nx,Ny,nmfgs);

		    for (y=0; y<Ny; y++) {
			for (x=0; x<Nx; x++)
			    GS[z][y][x]=(x<nx && y<ny)?cell[y][x]:0;

			if (y<ny) free(cell[y]);
		    }
		    free(cell);
		}
	    }

	    while (scanf("%d %d",&lower,&upper)==2)
		if (GS2BW(Nx,Ny,Nz,GS,lower,upper,BW,&X1,&Y1,&Z1,&X2,&Y2,&Z2)) {
		    sprintf(title,"%d\t%d\t",lower,upper);
		    CountCores(X1,Y1,Z1,X2,Y2,Z2,BW,last,title);
		}
	}
	return 0;
}
