
#include <stdio.h>
#include <stdlib.h>
#include "cell.h"
#include "rif.h"
#include "sif.h"
#include "rcm.h"
#include "sif_cm.h"

extern void	Error(/* char *msg */);

#ifndef	DILATIONS
#define DILATIONS	2
#endif

#define BITS	(sizeof(Cell)*8)
#define CORE	((1<<BITS)-1)

static int	Binarization(Nx,Ny,cell,lower,upper,X1,Y1,X2,Y2)
int		Nx,Ny,lower,upper,*X1,*Y1,*X2,*Y2;
Cell		**cell;
{
	int	y,x,pixels=0;

	for (y=0; y<Ny; y++)
	for (x=0; x<Nx; x++)
	    if (cell[y][x]>=lower && cell[y][x]<=upper) {
		cell[y][x]=CORE;
		if (++pixels==1) {
		    *X1=(*X2)=x;
		    *Y1=(*Y2)=y;
		}
		else {
		    if (x<*X1) *X1=x; else if (x>*X2) *X2=x;
		    if (y<*Y1) *Y1=y; else if (y>*Y2) *Y2=y;
		}
	    }
	    else
		cell[y][x]=0;

	return pixels;
}

static int	Erosion(x1,y1,x2,y2,cell,last,layer)
int		*x1,*y1,*x2,*y2,*last,layer;
Cell		**cell;
{
	int	x,y,X1=(*x1),Y1=(*y1),X2=(*x2),Y2=(*y2),pixels=0;

	for (x=X1-1; x<=X2; x++) last[x]=0;

	for (y=Y1; y<=Y2; y++)
	for (x=X1; x<=X2; x++)
	    if (cell[y][x]==CORE) {
		if (last[x]!=CORE ||
		    last[x-1]!=CORE ||
		    y==Y2 || cell[y+1][x]!=CORE ||
		    x==X2 || cell[y][x+1]!=CORE)
		    cell[y][x]=layer;
		else
		    if (++pixels==1) {
			*x1=(*x2)=x;
			*y1=(*y2)=y;
		    }
		    else {
			if (x<*x1) *x1=x; else if (x>*x2) *x2=x;
			if (y<*y1) *y1=y; else if (y>*y2) *y2=y;
		    }

		last[x]=CORE;
	    }
	    else
		last[x]=cell[y][x];

	return pixels;
}

static int	Search(table,label)
int		*table,label;
{
	while (table[label]<0) label=(-table[label]);

	return label;
}

#define IA(noe)	(int *)malloc((size_t)(noe)*sizeof(int))

static int	Labeling(x1,y1,x2,y2,cell,last)
int		x1,y1,x2,y2,*last;
Cell		**cell;
{
	int	x,y,*table,label,new,old,pvs,labels=0;

	for (x=x1-1; x<=x2; x++) last[x]=0;

	for (y=y1; y<=y2; y++)
	for (x=x1; x<=x2; x++) {
	    if (cell[y][x]==CORE && last[x]!=CORE && last[x-1]!=CORE) ++labels;

	    last[x]=cell[y][x];
	}

	if ((table=IA(labels))==NULL) Error("no memory for 'table'.");

	--table;

	label=1;
	for (x=x1; x<=x2; x++) last[x]=0;
	for (y=y1; y<=y2; y++)
	for (x=x1; x<=x2; x++)
	    if (cell[y][x]!=CORE)
		last[x]=0;
	    else {
		if (last[x]) {
		    new=Search(table,last[x]);
		    if (last[x-1] && (old=last[x-1])!=new) table[old]=(-new);
		}
		else if (last[x-1])
		    new=last[x-1];
		else
		    table[new=label++]=0;

		last[x]=new;
	    }

	pvs=2;
	for (label=1; label<=labels; label++)
	    if (!table[label]) table[label]=pvs++;

	if (pvs>>BITS) Error("too many cores.");

	for (label=1; label<=labels; label++)
	    if (table[label]<0) table[label]=table[Search(table,label)];

	label=1;
	for (x=x1; x<=x2; x++) last[x]=0;
	for (y=y1; y<=y2; y++)
	for (x=x1; x<=x2; x++)
	    if (cell[y][x]!=CORE)
		last[x]=0;
	    else
		if (last[x])
		    cell[y][x]=last[x];
		else if (last[x-1])
		    cell[y][x]=last[x]=last[x-1];
		else
		    cell[y][x]=last[x]=table[label++];

	free(table+1); return pvs;
}

#define BYTE	unsigned char

static void	Dilation(X1,Y1,X2,Y2,cell,last,flag,layer)
int		X1,Y1,X2,Y2,*last,layer;
Cell		**cell;
BYTE		**flag;
{
	int		pixels,x,y,pv,pv0,pv1,pv2,pv3;
	BYTE		*ptr1,bit1,*ptr0,bit0,*ptr,bit;
	unsigned int	dilations=(unsigned int)DILATIONS;

	last[X1-1]=0;

	do {
	    pixels=0;

	    for (x=X1; x<=X2; x++) last[x]=0;

	    ptr1=flag[Y1/8];
	    bit1=(BYTE)(1<<(Y1%8));
	    for (y=Y1; y<=Y2; y++) {
		ptr0=ptr1; ptr1=flag[(y+1)/8];
		bit0=bit1; bit1=(BYTE)(1<<((y+1)%8));

		for (x=X1; x<=X2; x++)
		    if (!(ptr0[x]&bit0)) {
			if (cell[y][x]==layer) {
			    pv=pv0=last[x];
			    if ((pv1=last[x-1])>pv) pv=pv1;

		if ((pv2=(y==Y2)?0:(ptr1[x]&bit1)?cell[y+1][x]:0)>pv) pv=pv2;
		if ((pv3=(x==X2)?0:(ptr0[x+1]&bit0)?cell[y][x+1]:0)>pv) pv=pv3;

			    if (pv>=2 && (pv0<=1 || pv0==pv)
				      && (pv1<=1 || pv1==pv)
				      && (pv2<=1 || pv2==pv)
				      && (pv3<=1 || pv3==pv)) {
				ptr0[x]|=bit0; cell[y][x]=pv; ++pixels;
			    }
			}
			last[x]=1;
		    }
		    else
			last[x]=cell[y][x];
	    }
	} while (--dilations && pixels);

	for (y=Y1; y<=Y2; y++) {
	    ptr=flag[y/8]; bit=(BYTE)(1<<(y%8));
	    for (x=X1; x<=X2; x++)
		if (!(ptr[x]&bit) && cell[y][x]==layer) {
		    ptr[x]|=bit; cell[y][x]=1;
		}
	}
}

static void	Marking(X1,Y1,X2,Y2,cell,last)
int		X1,Y1,X2,Y2,*last;
Cell		**cell;
{
	int	x,y;
	Cell	pv;

	for (x=X1-1; x<=X2; x++) last[x]=0;

	for (y=Y1; y<=Y2; y++)
	for (x=X1; x<=X2; x++) {
	    if ((pv=cell[y][x])>=2 &&
		((last[x]>=2 && last[x]!=pv) ||
		 (last[x-1]>=2 && last[x-1]!=pv) ||
		 (y<Y2 && cell[y+1][x]>=2 && cell[y+1][x]!=pv) ||
		 (x<X2 && cell[y][x+1]>=2 && cell[y][x+1]!=pv))) cell[y][x]=1;

	    last[x]=pv;
	}
}

#define MA(noe,ptr)	malloc((size_t)(noe)*sizeof(*(ptr)))

int	main(argc,argv)
int	argc;
char	**argv;
{
	int		Nx,Ny,X1,Y1,X2,Y2,*last,
			x1,y1,x2,y2,layers,als,layer,y,x,pvs,bps,cms=0;
	unsigned short	*R,*G,*B;
	Cell		**cell;
	BYTE		**flag,*ptr,bit;
	char		*nmff="no memory for 'flag'.";

	if (argc!=6 && argc!=7)
	    Error("usage : gi_2d orgTIFF lower upper layers {CM} newTIFF");

	if (argc==7) cms=ReadCM(argv[5],&R,&G,&B);

	ReadImageFile(argv[1],&Nx,&Ny,NULL,&cell,NULL);

	if (!Binarization(Nx,Ny,cell,atoi(argv[2]),
				     atoi(argv[3]),&X1,&Y1,&X2,&Y2))
	    Error("no object pixel.");

	if ((last=IA(X2-X1+2))==NULL) Error("no memory for 'last'.");

	last-=(X1-1);

	x1=X1; y1=Y1; x2=X2; y2=Y2;

	if ((als=abs(layers=atoi(argv[4])))>=CORE) Error("too many layers.");

	for (layer=1; layer<=als; layer++)
	    if (!Erosion(&x1,&y1,&x2,&y2,cell,last,layer))
		Error("no core pixel.");

	if (layers>0) {
	    if ((flag=(BYTE **)MA(Y2/8-Y1/8+1,flag))==NULL) Error(nmff);

	    flag-=(Y1/8);

	    for (y=Y1/8; y<=Y2/8; y++) {
		if ((flag[y]=(BYTE *)MA(X2-X1+1,*flag))==NULL) Error(nmff);

		flag[y]-=X1;

		for (x=X1; x<=X2; x++) flag[y][x]=0;
	    }
	    for (y=y1; y<=y2; y++) {
		ptr=flag[y/8]; bit=(BYTE)(1<<(y%8));
		for (x=x1; x<=x2; x++) if (cell[y][x]==CORE) ptr[x]|=bit;
	    }
	}
	else
	    for (y=Y1; y<=Y2; y++)
	    for (x=X1; x<=X2; x++)
		if (cell[y][x] && cell[y][x]!=CORE) cell[y][x]=1;

	printf("%d\n",pvs=Labeling(x1,y1,x2,y2,cell,last));

	if (layers>0) {
	    for (layer=layers; layer>0; layer--)
		Dilation(X1,Y1,X2,Y2,cell,last,flag,layer);

	    if (argv[4][0]=='+') Marking(X1,Y1,X2,Y2,cell,last);
	}

	bps=(pvs<=4)?2:(pvs<=16)?4:(pvs<=256)?8:16;

	if (cms)
	    if (cms<1<<bps) {
		fputs("too few colormap (warning).\n",stderr);

		StoreImageFile(argv[6],Nx,Ny,bps,cell,NULL);
	    }
	    else
		StoreImageFile_CM(argv[6],Nx,Ny,bps,R,G,B,cell,NULL);
	else
	    StoreImageFile(argv[5],Nx,Ny,bps,cell,NULL);

	return 0;
}
