
#include <stdio.h>
#include <stdlib.h>
#include "cell.h"
#include "rif.h"
#include "sif.h"
#include "rcm.h"
#include "sif_cm.h"

extern void	Error(/* char *msg */);

static int	Binarization(Nx,Ny,cell,lower,upper,X1,Y1,X2,Y2)
int		Nx,Ny,lower,upper,*X1,*Y1,*X2,*Y2;
Cell		**cell;
{
	int	y,x,pixels=0;

	for (y=0; y<Ny; y++)
	for (x=0; x<Nx; x++)
	    if (cell[y][x]>=lower && cell[y][x]<=upper) {
		cell[y][x]=2;
		if (++pixels==1) {
		    *X1=(*X2)=x;
		    *Y1=(*Y2)=y;
		}
		else {
		    if (x<*X1) *X1=x; else if (x>*X2) *X2=x;
		    if (y<*Y1) *Y1=y; else if (y>*Y2) *Y2=y;
		}
	    }
	    else
		cell[y][x]=0;

	return pixels;
}

static int	Erosion(x1,y1,x2,y2,cell,last)
int		*x1,*y1,*x2,*y2,*last;
Cell		**cell;
{
	int	x,y,X1=(*x1),Y1=(*y1),X2=(*x2),Y2=(*y2),pixels=0;

	for (x=X1-1; x<=X2; x++) last[x]=0;

	for (y=Y1; y<=Y2; y++)
	for (x=X1; x<=X2; x++)
	    if (cell[y][x]==2) {
		if (last[x]!=2 ||
		    last[x-1]!=2 ||
		    y==Y2 || cell[y+1][x]!=2 ||
		    x==X2 || cell[y][x+1]!=2)
		    cell[y][x]=1;
		else
		    if (++pixels==1) {
			*x1=(*x2)=x;
			*y1=(*y2)=y;
		    }
		    else {
			if (x<*x1) *x1=x; else if (x>*x2) *x2=x;
			if (y<*y1) *y1=y; else if (y>*y2) *y2=y;
		    }

		last[x]=2;
	    }
	    else
		last[x]=cell[y][x];

	return pixels;
}

static int	Search(table,label)
int		*table,label;
{
	while (table[label]<0) label=(-table[label]);

	return label;
}

#define IA(noe)	(int *)malloc((size_t)(noe)*sizeof(int))

static int	Labeling(x1,y1,x2,y2,cell,last)
int		x1,y1,x2,y2,*last;
Cell		**cell;
{
	int	x,y,*table,label,new,old,pvs,labels=0;

	for (x=x1-1; x<=x2; x++) last[x]=0;

	for (y=y1; y<=y2; y++)
	for (x=x1; x<=x2; x++) {
	    if (cell[y][x]==2 && last[x]!=2 && last[x-1]!=2) ++labels;

	    last[x]=cell[y][x];
	}

	if ((table=IA(labels))==NULL) Error("no memory for 'table'.");

	--table;

	label=1;
	for (x=x1; x<=x2; x++) last[x]=0;
	for (y=y1; y<=y2; y++)
	for (x=x1; x<=x2; x++)
	    if (cell[y][x]!=2)
		last[x]=0;
	    else {
		if (last[x]) {
		    new=Search(table,last[x]);
		    if (last[x-1] && (old=last[x-1])!=new) table[old]=(-new);
		}
		else if (last[x-1])
		    new=last[x-1];
		else
		    table[new=label++]=0;

		last[x]=new;
	    }

	pvs=2;
	for (label=1; label<=labels; label++)
	    if (!table[label]) table[label]=pvs++;

	if (pvs>>(sizeof(Cell)*8)) Error("too many cores.");

	for (label=1; label<=labels; label++)
	    if (table[label]<0) table[label]=table[Search(table,label)];

	label=1;
	for (x=x1; x<=x2; x++) last[x]=0;
	for (y=y1; y<=y2; y++)
	for (x=x1; x<=x2; x++)
	    if (cell[y][x]!=2)
		last[x]=0;
	    else
		if (last[x])
		    cell[y][x]=last[x];
		else if (last[x-1])
		    cell[y][x]=last[x]=last[x-1];
		else
		    cell[y][x]=last[x]=table[label++];

	free(table+1); return pvs;
}

static int	Dilation(X1,Y1,X2,Y2,cell,last)
int		X1,Y1,X2,Y2,*last;
Cell		**cell;
{
	int	x,y,pv,pv0,pv1,pv2,pv3,pixels=0;

	for (x=X1-1; x<=X2; x++) last[x]=0;

	for (y=Y1; y<=Y2; y++)
	for (x=X1; x<=X2; x++)
	    if (cell[y][x]==1) {
		pv=pv0=last[x];
		if ((pv1=last[x-1])>pv) pv=pv1;
		if ((pv2=(y==Y2)?0:cell[y+1][x])>pv) pv=pv2;
		if ((pv3=(x==X2)?0:cell[y][x+1])>pv) pv=pv3;
		if (pv>=2 && (pv0<=1 || pv0==pv)
			  && (pv1<=1 || pv1==pv)
			  && (pv2<=1 || pv2==pv)
			  && (pv3<=1 || pv3==pv)) {
		    cell[y][x]=pv; ++pixels;
		}
		last[x]=1;
	    }
	    else
		last[x]=cell[y][x];

	return pixels;
}

static void	Marking(X1,Y1,X2,Y2,cell,last)
int		X1,Y1,X2,Y2,*last;
Cell		**cell;
{
	int	x,y;
	Cell	pv;

	for (x=X1-1; x<=X2; x++) last[x]=0;

	for (y=Y1; y<=Y2; y++)
	for (x=X1; x<=X2; x++) {
	    if ((pv=cell[y][x])>=2 &&
		((last[x]>=2 && last[x]!=pv) ||
		 (last[x-1]>=2 && last[x-1]!=pv) ||
		 (y<Y2 && cell[y+1][x]>=2 && cell[y+1][x]!=pv) ||
		 (x<X2 && cell[y][x+1]>=2 && cell[y][x+1]!=pv))) cell[y][x]=1;

	    last[x]=pv;
	}
}

int	main(argc,argv)
int	argc;
char	**argv;
{
	int		Nx,Ny,X1,Y1,X2,Y2,
			*last,x1,y1,x2,y2,layers,layer,pvs,bps,cms=0;
	unsigned short	*R,*G,*B;
	Cell		**cell;

	if (argc!=6 && argc!=7)
	    Error("usage : gi_2d orgTIFF lower upper layers {CM} newTIFF");

	if (argc==7) cms=ReadCM(argv[5],&R,&G,&B);

	ReadImageFile(argv[1],&Nx,&Ny,NULL,&cell,NULL);

	if (!Binarization(Nx,Ny,cell,atoi(argv[2]),
				     atoi(argv[3]),&X1,&Y1,&X2,&Y2))
	    Error("no object pixel.");

	if ((last=IA(X2-X1+2))==NULL) Error("no memory for 'last'.");

	last-=(X1-1);

	x1=X1; y1=Y1; x2=X2; y2=Y2; layers=atoi(argv[4]);

	for (layer=abs(layers); layer>0; layer--)
	    if (!Erosion(&x1,&y1,&x2,&y2,cell,last)) Error("no core pixel.");

	printf("%d\n",pvs=Labeling(x1,y1,x2,y2,cell,last));

	if (layers>0) {
	    while (Dilation(X1,Y1,X2,Y2,cell,last)) ;

	    if (argv[4][0]=='+') Marking(X1,Y1,X2,Y2,cell,last);
	}

	bps=(pvs<=4)?2:(pvs<=16)?4:(pvs<=256)?8:16;

	if (cms)
	    if (cms<1<<bps) {
		fputs("too few colormap (warning).\n",stderr);

		StoreImageFile(argv[6],Nx,Ny,bps,cell,NULL);
	    }
	    else
		StoreImageFile_CM(argv[6],Nx,Ny,bps,R,G,B,cell,NULL);
	else
	    StoreImageFile(argv[5],Nx,Ny,bps,cell,NULL);

	return 0;
}
