
#include <stdio.h>
#include <stdlib.h>
#include "cell.h"
#include "rif.h"
#include "csi.h"
#include "sif.h"
#include "rcm.h"
#include "sif_cm.h"

extern void	Error(/* char *msg */);

#ifndef	DILATIONS
#define DILATIONS	3
#endif

#define BITS	(sizeof(Cell)*8)
#define CORE	((1<<BITS)-1)

static int	Binarization(Nx,Ny,Nz,cell,lower,upper,X1,Y1,Z1,X2,Y2,Z2)
int		Nx,Ny,Nz,lower,upper,*X1,*Y1,*Z1,*X2,*Y2,*Z2;
Cell		***cell;
{
	int	z,y,x,pixels=0;

	for (z=0; z<Nz; z++)
	for (y=0; y<Ny; y++)
	for (x=0; x<Nx; x++)
	    if (cell[z][y][x]>=lower && cell[z][y][x]<=upper) {
		cell[z][y][x]=CORE;
		if (++pixels==1) {
		    *X1=(*X2)=x;
		    *Y1=(*Y2)=y;
		    *Z1=(*Z2)=z;
		}
		else {
		    if (x<*X1) *X1=x; else if (x>*X2) *X2=x;
		    if (y<*Y1) *Y1=y; else if (y>*Y2) *Y2=y;
		    if (z<*Z1) *Z1=z; else if (z>*Z2) *Z2=z;
		}
	    }
	    else
		cell[z][y][x]=0;

	return pixels;
}

static int	Erosion(x1,y1,z1,x2,y2,z2,cell,last,layer)
int		*x1,*y1,*z1,*x2,*y2,*z2,**last,layer;
Cell		***cell;
{
	int	y,x,z,
		X1=(*x1),Y1=(*y1),Z1=(*z1),X2=(*x2),Y2=(*y2),Z2=(*z2),pixels=0;

	for (y=Y1-1; y<=Y2; y++)
	for (x=X1-1; x<=X2; x++) last[y][x]=0;

	for (z=Z1; z<=Z2; z++)
	for (y=Y1; y<=Y2; y++)
	for (x=X1; x<=X2; x++)
	    if (cell[z][y][x]==CORE) {
		if (last[y][x]!=CORE ||
		    last[y-1][x]!=CORE ||
		    last[y][x-1]!=CORE ||
		    z==Z2 || cell[z+1][y][x]!=CORE ||
		    y==Y2 || cell[z][y+1][x]!=CORE ||
		    x==X2 || cell[z][y][x+1]!=CORE)
		    cell[z][y][x]=layer;
		else
		    if (++pixels==1) {
			*x1=(*x2)=x;
			*y1=(*y2)=y;
			*z1=(*z2)=z;
		    }
		    else {
			if (x<*x1) *x1=x; else if (x>*x2) *x2=x;
			if (y<*y1) *y1=y; else if (y>*y2) *y2=y;
			if (z<*z1) *z1=z; else if (z>*z2) *z2=z;
		    }

		last[y][x]=CORE;
	    }
	    else
		last[y][x]=cell[z][y][x];

	return pixels;
}

static int	Search(table,label)
int		*table,label;
{
	while (table[label]<0) label=(-table[label]);

	return label;
}

#define IA(noe)	(int *)malloc((size_t)(noe)*sizeof(int))

static int	Labeling(x1,y1,z1,x2,y2,z2,cell,last)
int		x1,y1,z1,x2,y2,z2,**last;
Cell		***cell;
{
	int	y,x,z,*table,label,new,old,pvs,labels=0;

	for (y=y1-1; y<=y2; y++)
	for (x=x1-1; x<=x2; x++) last[y][x]=0;

	for (z=z1; z<=z2; z++)
	for (y=y1; y<=y2; y++)
	for (x=x1; x<=x2; x++) {
	    if (cell[z][y][x]==CORE && last[y][x]!=CORE &&
				       last[y-1][x]!=CORE &&
				       last[y][x-1]!=CORE) ++labels;

	    last[y][x]=cell[z][y][x];
	}

	if ((table=IA(labels))==NULL) Error("no memory for 'table'.");

	--table;

	label=1;
	for (y=y1; y<=y2; y++)
	for (x=x1; x<=x2; x++) last[y][x]=0;
	for (z=z1; z<=z2; z++)
	for (y=y1; y<=y2; y++)
	for (x=x1; x<=x2; x++)
	    if (cell[z][y][x]!=CORE)
		last[y][x]=0;
	    else {
		if (last[y][x]) {
		    new=Search(table,last[y][x]);
		    if (last[y-1][x] && (old=Search(table,last[y-1][x]))!=new)
			table[old]=(-new);
		    if (last[y][x-1] && (old=Search(table,last[y][x-1]))!=new)
			table[old]=(-new);
		}
		else if (last[y-1][x]) {
		    new=Search(table,last[y-1][x]);
		    if (last[y][x-1] && (old=last[y][x-1])!=new)
			table[old]=(-new);
		}
		else if (last[y][x-1])
		    new=last[y][x-1];
		else
		    table[new=label++]=0;

		last[y][x]=new;
	    }

	pvs=2;
	for (label=1; label<=labels; label++)
	    if (!table[label]) table[label]=pvs++;

	if (pvs>>BITS) Error("too many cores.");

	for (label=1; label<=labels; label++)
	    if (table[label]<0) table[label]=table[Search(table,label)];

	label=1;
	for (y=y1; y<=y2; y++)
	for (x=x1; x<=x2; x++) last[y][x]=0;
	for (z=z1; z<=z2; z++)
	for (y=y1; y<=y2; y++)
	for (x=x1; x<=x2; x++)
	    if (cell[z][y][x]!=CORE)
		last[y][x]=0;
	    else
		if (last[y][x])
		    cell[z][y][x]=last[y][x];
		else if (last[y-1][x])
		    cell[z][y][x]=last[y][x]=last[y-1][x];
		else if (last[y][x-1])
		    cell[z][y][x]=last[y][x]=last[y][x-1];
		else
		    cell[z][y][x]=last[y][x]=table[label++];

	free(table+1); return pvs;
}

#define BYTE	unsigned char

static void	Dilation(X1,Y1,Z1,X2,Y2,Z2,cell,last,flag,layer)
int		X1,Y1,Z1,X2,Y2,Z2,**last,layer;
Cell		***cell;
BYTE		***flag;
{
	int		x,y,pixels,z,pv,pv0,pv1,pv2,pv3,pv4,pv5;
	BYTE		**ptr1,bit1,**ptr0,bit0,**ptr,bit;
	unsigned int	dilations=(unsigned int)DILATIONS;

	last[Y1-1][X1-1]=0; x=X1-1; for (y=Y1; y<=Y2; y++) last[y][x]=0;
			    y=Y1-1; for (x=X1; x<=X2; x++) last[y][x]=0;

	do {
	    pixels=0;

	    for (y=Y1; y<=Y2; y++)
	    for (x=X1; x<=X2; x++) last[y][x]=0;

	    ptr1=flag[Z1/8];
	    bit1=(BYTE)(1<<(Z1%8));
	    for (z=Z1; z<=Z2; z++) {
		ptr0=ptr1; ptr1=flag[(z+1)/8];
		bit0=bit1; bit1=(BYTE)(1<<((z+1)%8));

		for (y=Y1; y<=Y2; y++)
		for (x=X1; x<=X2; x++)
		    if (!(ptr0[y][x]&bit0)) {
			if (cell[z][y][x]==layer) {
			    pv=pv0=last[y][x];
			    if ((pv1=last[y-1][x])>pv) pv=pv1;
			    if ((pv2=last[y][x-1])>pv) pv=pv2;

	if ((pv3=(z==Z2)?0:(ptr1[y][x]&bit1)?cell[z+1][y][x]:0)>pv) pv=pv3;
	if ((pv4=(y==Y2)?0:(ptr0[y+1][x]&bit0)?cell[z][y+1][x]:0)>pv) pv=pv4;
	if ((pv5=(x==X2)?0:(ptr0[y][x+1]&bit0)?cell[z][y][x+1]:0)>pv) pv=pv5;

			    if (pv>=2 && (pv0<=1 || pv0==pv)
				      && (pv1<=1 || pv1==pv)
				      && (pv2<=1 || pv2==pv)
				      && (pv3<=1 || pv3==pv)
				      && (pv4<=1 || pv4==pv)
				      && (pv5<=1 || pv5==pv)) {
				ptr0[y][x]|=bit0; cell[z][y][x]=pv; ++pixels;
			    }
			}
			last[y][x]=1;
		    }
		    else
			last[y][x]=cell[z][y][x];
	    }
	} while (--dilations && pixels);

	for (z=Z1; z<=Z2; z++) {
	    ptr=flag[z/8]; bit=(BYTE)(1<<(z%8));
	    for (y=Y1; y<=Y2; y++)
	    for (x=X1; x<=X2; x++)
		if (!(ptr[y][x]&bit) && cell[z][y][x]==layer) {
		    ptr[y][x]|=bit; cell[z][y][x]=1;
		}
	}
}

static void	Marking(X1,Y1,Z1,X2,Y2,Z2,cell,last)
int		X1,Y1,Z1,X2,Y2,Z2,**last;
Cell		***cell;
{
	int	y,x,z;
	Cell	pv;

	for (y=Y1-1; y<=Y2; y++)
	for (x=X1-1; x<=X2; x++) last[y][x]=0;

	for (z=Z1; z<=Z2; z++)
	for (y=Y1; y<=Y2; y++)
	for (x=X1; x<=X2; x++) {
	    if ((pv=cell[z][y][x])>=2 &&
		((last[y][x]>=2 && last[y][x]!=pv) ||
		 (last[y-1][x]>=2 && last[y-1][x]!=pv) ||
		 (last[y][x-1]>=2 && last[y][x-1]!=pv) ||
		 (z<Z2 && cell[z+1][y][x]>=2 && cell[z+1][y][x]!=pv) ||
		 (y<Y2 && cell[z][y+1][x]>=2 && cell[z][y+1][x]!=pv) ||
		 (x<X2 && cell[z][y][x+1]>=2 && cell[z][y][x+1]!=pv)))
		cell[z][y][x]=1;

	    last[y][x]=pv;
	}
}

#define MA(noe,ptr)	malloc((size_t)(noe)*sizeof(*(ptr)))

#define NAME_LEN	256

int	main(argc,argv)
int	argc;
char	**argv;
{
	int		Nx,Ny,Nz,z,nx,ny,y,x,X1,Y1,Z1,X2,Y2,Z2,**last,
			x1,y1,z1,x2,y2,z2,layers,als,layer,pvs,len,bps,cms=0;
	unsigned short	*R,*G,*B;
	char		**path,name[NAME_LEN],
			*nmfc="no memory for 'cell'.",
			*nmfl="no memory for 'last'.",
			*nmff="no memory for 'flag'.";
	Cell		***cell,**slice;
	unsigned char	***flag,**ptr,bit;

	if (argc!=7 && argc!=8)
	Error("usage : gi_3d orgDir nameFile lower upper layers {CM} newDir");

	if (argc==8) cms=ReadCM(argv[6],&R,&G,&B);

	path=CheckSliceImages(argv[1],argv[2],&Nx,&Ny,&Nz,NULL);

	if ((cell=(Cell ***)MA(Nz,cell))==NULL) Error(nmfc);

	for (z=0; z<Nz; z++) {
	    ReadImageFile(path[z],&nx,&ny,NULL,&slice,NULL);

	    if (nx==Nx && nx==Ny)
		cell[z]=slice;
	    else {
		if ((cell[z]=(Cell **)MA(Ny,*cell))==NULL) Error(nmfc);

		for (y=0; y<Ny; y++) {
		    if ((cell[z][y]=(Cell *)MA(Nx,**cell))==NULL) Error(nmfc);

		    for (x=0; x<Nx; x++)
			cell[z][y][x]=(x<nx && y<ny)?slice[y][x]:0;

		    if (y<ny) free(slice[y]);
		}
		free(slice);
	    }
	}

	if (!Binarization(Nx,Ny,Nz,cell,atoi(argv[3]),
					atoi(argv[4]),&X1,&Y1,&Z1,&X2,&Y2,&Z2))
	    Error("no object pixel.");

	if ((last=(int **)MA(Y2-Y1+2,last))==NULL) Error(nmfl);

	last-=(Y1-1);
	for (y=Y1-1; y<=Y2; y++) {
	    if ((last[y]=IA(X2-X1+2))==NULL) Error(nmfl);

	    last[y]-=(X1-1);
	}

	x1=X1; y1=Y1; z1=Z1; x2=X2; y2=Y2; z2=Z2;

	if ((als=abs(layers=atoi(argv[5])))>=CORE) Error("too many layers.");

	for (layer=1; layer<=als; layer++)
	    if (!Erosion(&x1,&y1,&z1,&x2,&y2,&z2,cell,last,layer))
		Error("no core pixel.");

	if (layers>0) {
	    if ((flag=(BYTE ***)MA(Z2/8-Z1/8+1,flag))==NULL) Error(nmff);

	    flag-=(Z1/8);

	    for (z=Z1/8; z<=Z2/8; z++) {
		if ((flag[z]=(BYTE **)MA(Y2-Y1+1,*flag))==NULL) Error(nmff);

		flag[z]-=Y1;

		for (y=Y1; y<=Y2; y++) {
		    if ((flag[z][y]=(BYTE *)MA(X2-X1+1,**flag))==NULL)
			Error(nmff);

		    flag[z][y]-=X1;

		    for (x=X1; x<=X2; x++) flag[z][y][x]=0;
		}
	    }
	    for (z=z1; z<=z2; z++) {
		ptr=flag[z/8]; bit=(BYTE)(1<<(z%8));
		for (y=y1; y<=y2; y++)
		for (x=x1; x<=x2; x++) if (cell[z][y][x]==CORE) ptr[y][x]|=bit;
	    }
	}
	else
	    for (z=Z1; z<=Z2; z++)
	    for (y=Y1; y<=Y2; y++)
	    for (x=X1; x<=X2; x++)
		if (cell[z][y][x] && cell[z][y][x]!=CORE) cell[z][y][x]=1;

	printf("%d\n",pvs=Labeling(x1,y1,z1,x2,y2,z2,cell,last));

	if (layers>0) {
	    for (layer=layers; layer>0; layer--)
		Dilation(X1,Y1,Z1,X2,Y2,Z2,cell,last,flag,layer);

	    if (argv[5][0]=='+') Marking(X1,Y1,Z1,X2,Y2,Z2,cell,last);
	}

	for (len=0; argv[1][len]!='\0'; len++) ;

	bps=(pvs<=4)?2:(pvs<=16)?4:(pvs<=256)?8:16;

	if (cms && cms<1<<bps) {
	    fputs("too few colormap (warning).\n",stderr);

	    cms=0;
	}
	for (z=0; z<Nz; z++) {
	    sprintf(name,"%s%s",argv[argc-1],path[z]+len);
	    if (cms)
		StoreImageFile_CM(name,Nx,Ny,bps,R,G,B,cell[z],NULL);
	    else
		StoreImageFile(name,Nx,Ny,bps,cell[z],NULL);
	}
	return 0;
}
