
#include <stdio.h>
#include <stdlib.h>
#include "cell.h"
#include "rif.h"
#include "csi.h"
#include "sif.h"
#include "rcm.h"
#include "sif_cm.h"

extern void	Error(/* char *msg */);

static int	Binarization(Nx,Ny,Nz,cell,lower,upper,X1,Y1,Z1,X2,Y2,Z2)
int		Nx,Ny,Nz,lower,upper,*X1,*Y1,*Z1,*X2,*Y2,*Z2;
Cell		***cell;
{
	int	z,y,x,pixels=0;

	for (z=0; z<Nz; z++)
	for (y=0; y<Ny; y++)
	for (x=0; x<Nx; x++)
	    if (cell[z][y][x]>=lower && cell[z][y][x]<=upper) {
		cell[z][y][x]=2;
		if (++pixels==1) {
		    *X1=(*X2)=x;
		    *Y1=(*Y2)=y;
		    *Z1=(*Z2)=z;
		}
		else {
		    if (x<*X1) *X1=x; else if (x>*X2) *X2=x;
		    if (y<*Y1) *Y1=y; else if (y>*Y2) *Y2=y;
		    if (z<*Z1) *Z1=z; else if (z>*Z2) *Z2=z;
		}
	    }
	    else
		cell[z][y][x]=0;

	return pixels;
}

static int	Erosion(x1,y1,z1,x2,y2,z2,cell,last)
int		*x1,*y1,*z1,*x2,*y2,*z2,**last;
Cell		***cell;
{
	int	y,x,z,
		X1=(*x1),Y1=(*y1),Z1=(*z1),X2=(*x2),Y2=(*y2),Z2=(*z2),pixels=0;

	for (y=Y1-1; y<=Y2; y++)
	for (x=X1-1; x<=X2; x++) last[y][x]=0;

	for (z=Z1; z<=Z2; z++)
	for (y=Y1; y<=Y2; y++)
	for (x=X1; x<=X2; x++)
	    if (cell[z][y][x]==2) {
		if (last[y][x]!=2 ||
		    last[y-1][x]!=2 ||
		    last[y][x-1]!=2 ||
		    z==Z2 || cell[z+1][y][x]!=2 ||
		    y==Y2 || cell[z][y+1][x]!=2 ||
		    x==X2 || cell[z][y][x+1]!=2)
		    cell[z][y][x]=1;
		else
		    if (++pixels==1) {
			*x1=(*x2)=x;
			*y1=(*y2)=y;
			*z1=(*z2)=z;
		    }
		    else {
			if (x<*x1) *x1=x; else if (x>*x2) *x2=x;
			if (y<*y1) *y1=y; else if (y>*y2) *y2=y;
			if (z<*z1) *z1=z; else if (z>*z2) *z2=z;
		    }

		last[y][x]=2;
	    }
	    else
		last[y][x]=cell[z][y][x];

	return pixels;
}

static int	Search(table,label)
int		*table,label;
{
	while (table[label]<0) label=(-table[label]);

	return label;
}

#define IA(noe)	(int *)malloc((size_t)(noe)*sizeof(int))

static int	Labeling(x1,y1,z1,x2,y2,z2,cell,last)
int		x1,y1,z1,x2,y2,z2,**last;
Cell		***cell;
{
	int	y,x,z,*table,label,new,old,pvs,labels=0;

	for (y=y1-1; y<=y2; y++)
	for (x=x1-1; x<=x2; x++) last[y][x]=0;

	for (z=z1; z<=z2; z++)
	for (y=y1; y<=y2; y++)
	for (x=x1; x<=x2; x++) {
	    if (cell[z][y][x]==2 &&
		last[y][x]!=2 && last[y-1][x]!=2 && last[y][x-1]!=2) ++labels;

	    last[y][x]=cell[z][y][x];
	}

	if ((table=IA(labels))==NULL) Error("no memory for 'table'.");

	--table;

	label=1;
	for (y=y1; y<=y2; y++)
	for (x=x1; x<=x2; x++) last[y][x]=0;
	for (z=z1; z<=z2; z++)
	for (y=y1; y<=y2; y++)
	for (x=x1; x<=x2; x++)
	    if (cell[z][y][x]!=2)
		last[y][x]=0;
	    else {
		if (last[y][x]) {
		    new=Search(table,last[y][x]);
		    if (last[y-1][x] && (old=Search(table,last[y-1][x]))!=new)
			table[old]=(-new);
		    if (last[y][x-1] && (old=Search(table,last[y][x-1]))!=new)
			table[old]=(-new);
		}
		else if (last[y-1][x]) {
		    new=Search(table,last[y-1][x]);
		    if (last[y][x-1] && (old=last[y][x-1])!=new)
			table[old]=(-new);
		}
		else if (last[y][x-1])
		    new=last[y][x-1];
		else
		    table[new=label++]=0;

		last[y][x]=new;
	    }

	pvs=2;
	for (label=1; label<=labels; label++)
	    if (!table[label]) table[label]=pvs++;

	if (pvs>>(sizeof(Cell)*8)) Error("too many cores.");

	for (label=1; label<=labels; label++)
	    if (table[label]<0) table[label]=table[Search(table,label)];

	label=1;
	for (y=y1; y<=y2; y++)
	for (x=x1; x<=x2; x++) last[y][x]=0;
	for (z=z1; z<=z2; z++)
	for (y=y1; y<=y2; y++)
	for (x=x1; x<=x2; x++)
	    if (cell[z][y][x]!=2)
		last[y][x]=0;
	    else
		if (last[y][x])
		    cell[z][y][x]=last[y][x];
		else if (last[y-1][x])
		    cell[z][y][x]=last[y][x]=last[y-1][x];
		else if (last[y][x-1])
		    cell[z][y][x]=last[y][x]=last[y][x-1];
		else
		    cell[z][y][x]=last[y][x]=table[label++];

	free(table+1); return pvs;
}

static int	Dilation(X1,Y1,Z1,X2,Y2,Z2,cell,last)
int		X1,Y1,Z1,X2,Y2,Z2,**last;
Cell		***cell;
{
	int	y,x,z,pv,pv0,pv1,pv2,pv3,pv4,pv5,pixels=0;

	for (y=Y1-1; y<=Y2; y++)
	for (x=X1-1; x<=X2; x++) last[y][x]=0;

	for (z=Z1; z<=Z2; z++)
	for (y=Y1; y<=Y2; y++)
	for (x=X1; x<=X2; x++)
	    if (cell[z][y][x]==1) {
		pv=pv0=last[y][x];
		if ((pv1=last[y-1][x])>pv) pv=pv1;
		if ((pv2=last[y][x-1])>pv) pv=pv2;
		if ((pv3=(z==Z2)?0:cell[z+1][y][x])>pv) pv=pv3;
		if ((pv4=(y==Y2)?0:cell[z][y+1][x])>pv) pv=pv4;
		if ((pv5=(x==X2)?0:cell[z][y][x+1])>pv) pv=pv5;
		if (pv>=2 && (pv0<=1 || pv0==pv)
			  && (pv1<=1 || pv1==pv)
			  && (pv2<=1 || pv2==pv)
			  && (pv3<=1 || pv3==pv)
			  && (pv4<=1 || pv4==pv)
			  && (pv5<=1 || pv5==pv)) {
		    cell[z][y][x]=pv; ++pixels;
		}
		last[y][x]=1;
	    }
	    else
		last[y][x]=cell[z][y][x];

	return pixels;
}

static void	Marking(X1,Y1,Z1,X2,Y2,Z2,cell,last)
int		X1,Y1,Z1,X2,Y2,Z2,**last;
Cell		***cell;
{
	int	y,x,z;
	Cell	pv;

	for (y=Y1-1; y<=Y2; y++)
	for (x=X1-1; x<=X2; x++) last[y][x]=0;

	for (z=Z1; z<=Z2; z++)
	for (y=Y1; y<=Y2; y++)
	for (x=X1; x<=X2; x++) {
	    if ((pv=cell[z][y][x])>=2 &&
		((last[y][x]>=2 && last[y][x]!=pv) ||
		 (last[y-1][x]>=2 && last[y-1][x]!=pv) ||
		 (last[y][x-1]>=2 && last[y][x-1]!=pv) ||
		 (z<Z2 && cell[z+1][y][x]>=2 && cell[z+1][y][x]!=pv) ||
		 (y<Y2 && cell[z][y+1][x]>=2 && cell[z][y+1][x]!=pv) ||
		 (x<X2 && cell[z][y][x+1]>=2 && cell[z][y][x+1]!=pv)))
		cell[z][y][x]=1;

	    last[y][x]=pv;
	}
}

#define MA(noe,ptr)	malloc((size_t)(noe)*sizeof(*(ptr)))

#define NAME_LEN	256

int	main(argc,argv)
int	argc;
char	**argv;
{
	int		Nx,Ny,Nz,z,nx,ny,y,x,X1,Y1,Z1,X2,Y2,Z2,
			**last,x1,y1,z1,x2,y2,z2,layers,layer,pvs,len,bps,cms=0;
	unsigned short	*R,*G,*B;
	char		**path,name[NAME_LEN],
			*nmfc="no memory for 'cell'.",
			*nmfl="no memory for 'last'.";
	Cell		***cell,**slice;

	if (argc!=7 && argc!=8)
	Error("usage : gi_3d orgDir nameFile lower upper layers {CM} newDir");

	if (argc==8) cms=ReadCM(argv[6],&R,&G,&B);

	path=CheckSliceImages(argv[1],argv[2],&Nx,&Ny,&Nz,NULL);

	if ((cell=(Cell ***)MA(Nz,cell))==NULL) Error(nmfc);

	for (z=0; z<Nz; z++) {
	    ReadImageFile(path[z],&nx,&ny,NULL,&slice,NULL);

	    if (nx==Nx && nx==Ny)
		cell[z]=slice;
	    else {
		if ((cell[z]=(Cell **)MA(Ny,*cell))==NULL) Error(nmfc);

		for (y=0; y<Ny; y++) {
		    if ((cell[z][y]=(Cell *)MA(Nx,**cell))==NULL) Error(nmfc);

		    for (x=0; x<Nx; x++)
			cell[z][y][x]=(x<nx && y<ny)?slice[y][x]:0;

		    if (y<ny) free(slice[y]);
		}
		free(slice);
	    }
	}

	if (!Binarization(Nx,Ny,Nz,cell,atoi(argv[3]),
					atoi(argv[4]),&X1,&Y1,&Z1,&X2,&Y2,&Z2))
	    Error("no object pixel.");

	if ((last=(int **)MA(Y2-Y1+2,last))==NULL) Error(nmfl);

	last-=(Y1-1);
	for (y=Y1-1; y<=Y2; y++) {
	    if ((last[y]=IA(X2-X1+2))==NULL) Error(nmfl);

	    last[y]-=(X1-1);
	}

	x1=X1; y1=Y1; z1=Z1; x2=X2; y2=Y2; z2=Z2; layers=atoi(argv[5]);

	for (layer=abs(layers); layer>0; layer--)
	    if (!Erosion(&x1,&y1,&z1,&x2,&y2,&z2,cell,last))
		Error("no core pixel.");

	printf("%d\n",pvs=Labeling(x1,y1,z1,x2,y2,z2,cell,last));

	if (layers>0) {
	    while (Dilation(X1,Y1,Z1,X2,Y2,Z2,cell,last)) ;

	    if (argv[5][0]=='+') Marking(X1,Y1,Z1,X2,Y2,Z2,cell,last);
	}

	for (len=0; argv[1][len]!='\0'; len++) ;

	bps=(pvs<=4)?2:(pvs<=16)?4:(pvs<=256)?8:16;

	if (cms && cms<1<<bps) {
	    fputs("too few colormap (warning).\n",stderr);

	    cms=0;
	}
	for (z=0; z<Nz; z++) {
	    sprintf(name,"%s%s",argv[argc-1],path[z]+len);
	    if (cms)
		StoreImageFile_CM(name,Nx,Ny,bps,R,G,B,cell[z],NULL);
	    else
		StoreImageFile(name,Nx,Ny,bps,cell[z],NULL);
	}
	return 0;
}
