
#include <stdio.h>
#include <stdlib.h>
#include "cell.h"
#include "rif.h"
#include "sif.h"
#include "rcm.h"
#include "sif_cm.h"

extern void	Error(/* char *msg */);

#define CORE	((1<<(sizeof(Cell)*8))-1)

static int	Binarization(Nx,Ny,cell,lower,upper,x1,y1,x2,y2)
int		Nx,Ny,lower,upper,*x1,*y1,*x2,*y2;
Cell		**cell;
{
	int	y,x,pixels=0;

	for (y=0; y<Ny; y++)
	for (x=0; x<Nx; x++)
	    if (cell[y][x]>=lower && cell[y][x]<=upper) {
		cell[y][x]=CORE;
		if (++pixels==1) {
		    *x1=(*x2)=x;
		    *y1=(*y2)=y;
		}
		else {
		    if (x<*x1) *x1=x; else if (x>*x2) *x2=x;
		    if (y<*y1) *y1=y; else if (y>*y2) *y2=y;
		}
	    }
	    else
		cell[y][x]=0;

	return pixels;
}

static int	Erosion(x1,y1,x2,y2,cell,last,layer)
int		*x1,*y1,*x2,*y2,layer;
Cell		**cell,*last;
{
	int	y,x,X1=(*x1),Y1=(*y1),X2=(*x2),Y2=(*y2),pixels=0;

	for (y=Y1; y<=Y2; y++)
	for (x=X1; x<=X2; x++)
	    if (cell[y][x]==CORE) {
		if (y==Y1 || last[x]     !=CORE ||
		    x==X1 || last[x-1]   !=CORE ||
		    y==Y2 || cell[y+1][x]!=CORE ||
		    x==X2 || cell[y][x+1]!=CORE)
		    cell[y][x]=layer;
		else
		    if (++pixels==1) {
			*x1=(*x2)=x;
			*y1=(*y2)=y;
		    }
		    else {
			if (x<*x1) *x1=x; else if (x>*x2) *x2=x;
			if (y<*y1) *y1=y; else if (y>*y2) *y2=y;
		    }

		last[x]=CORE;
	    }
	    else
		last[x]=cell[y][x];

	return pixels;
}

int	main(argc,argv)
int	argc;
char	**argv;
{
	int		Nx,Ny,x1,y1,x2,y2,layer,bps,cms=0;
	unsigned short	*R,*G,*B;
	Cell		**cell,*last;

	if (argc!=4 && argc!=5 && argc!=6)
	    Error("usage : lm_2d orgTIFF lower upper {{CM} newTIFF}");

	if (argc==6) cms=ReadCM(argv[4],&R,&G,&B);

	ReadImageFile(argv[1],&Nx,&Ny,NULL,&cell,NULL);

	if (!Binarization(Nx,Ny,cell,atoi(argv[2]),
				     atoi(argv[3]),&x1,&y1,&x2,&y2))
	    Error("no object pixel.");

	if ((last=(Cell *)malloc((x2-x1+1)*sizeof(*last)))==NULL)
	    Error("no memory for 'last'.");

	last-=x1;

	for (layer=1; layer<CORE; layer++)
	    if (!Erosion(&x1,&y1,&x2,&y2,cell,last,layer)) break;

	printf("%d\n",layer);

	if (argc==5 || argc==6) {
	    bps=(layer<2)?1:(layer<4)?2:(layer<16)?4:(layer<256)?8:16;

	    if (cms)
		if (cms<1<<bps) {
		    fputs("too few colormap (warning).\n",stderr);

		    StoreImageFile(argv[5],Nx,Ny,bps,cell,NULL);
		}
		else
		    StoreImageFile_CM(argv[5],Nx,Ny,bps,R,G,B,cell,NULL);
	    else
		StoreImageFile(argv[4],Nx,Ny,bps,cell,NULL);
	}
	return 0;
}
