
#include <stdio.h>
#include <stdlib.h>
#include "cell.h"
#include "rif.h"
#include "csi.h"
#include "sif.h"
#include "rcm.h"
#include "sif_cm.h"

extern void	Error(/* char *msg */);

#define CORE	((1<<(sizeof(Cell)*8))-1)

static int	Binarization(Nx,Ny,Nz,cell,lower,upper,x1,y1,z1,x2,y2,z2)
int		Nx,Ny,Nz,lower,upper,*x1,*y1,*z1,*x2,*y2,*z2;
Cell		***cell;
{
	int	z,y,x,pixels=0;

	for (z=0; z<Nz; z++)
	for (y=0; y<Ny; y++)
	for (x=0; x<Nx; x++)
	    if (cell[z][y][x]>=lower && cell[z][y][x]<=upper) {
		cell[z][y][x]=CORE;
		if (++pixels==1) {
		    *x1=(*x2)=x;
		    *y1=(*y2)=y;
		    *z1=(*z2)=z;
		}
		else {
		    if (x<*x1) *x1=x; else if (x>*x2) *x2=x;
		    if (y<*y1) *y1=y; else if (y>*y2) *y2=y;
		    if (z<*z1) *z1=z; else if (z>*z2) *z2=z;
		}
	    }
	    else
		cell[z][y][x]=0;

	return pixels;
}

static int	Erosion(x1,y1,z1,x2,y2,z2,cell,last,layer)
int		*x1,*y1,*z1,*x2,*y2,*z2,layer;
Cell		***cell,**last;
{
	int	z,y,x,
		X1=(*x1),Y1=(*y1),Z1=(*z1),X2=(*x2),Y2=(*y2),Z2=(*z2),pixels=0;

	for (z=Z1; z<=Z2; z++)
	for (y=Y1; y<=Y2; y++)
	for (x=X1; x<=X2; x++)
	    if (cell[z][y][x]==CORE) {
		if (z==Z1 || last[y][x]     !=CORE ||
		    y==Y1 || last[y-1][x]   !=CORE ||
		    x==X1 || last[y][x-1]   !=CORE ||
		    z==Z2 || cell[z+1][y][x]!=CORE ||
		    y==Y2 || cell[z][y+1][x]!=CORE ||
		    x==X2 || cell[z][y][x+1]!=CORE)
		    cell[z][y][x]=layer;
		else
		    if (++pixels==1) {
			*x1=(*x2)=x;
			*y1=(*y2)=y;
			*z1=(*z2)=z;
		    }
		    else {
			if (x<*x1) *x1=x; else if (x>*x2) *x2=x;
			if (y<*y1) *y1=y; else if (y>*y2) *y2=y;
			if (z<*z1) *z1=z; else if (z>*z2) *z2=z;
		    }

		last[y][x]=CORE;
	    }
	    else
		last[y][x]=cell[z][y][x];

	return pixels;
}

#define MA(noe,ptr)	malloc((size_t)(noe)*sizeof(*(ptr)))

#define NAME_LEN	256

int	main(argc,argv)
int	argc;
char	**argv;
{
	int		Nx,Ny,Nz,z,nx,ny,y,x,
			x1,y1,z1,x2,y2,z2,layer,len,bps,cms=0;
	unsigned short	*R,*G,*B;
	char		**path,name[NAME_LEN],
			*nmfc="no memory for 'cell'.",
			*nmfl="no memory for 'last'.";
	Cell		***cell,**slice,**last;

	if (argc!=5 && argc!=6 && argc!=7)
	    Error("usage : lm_3d orgDir nameFile lower upper {{CM} newDir}");

	if (argc==7) cms=ReadCM(argv[5],&R,&G,&B);

	path=CheckSliceImages(argv[1],argv[2],&Nx,&Ny,&Nz,NULL);

	if ((cell=(Cell ***)MA(Nz,cell))==NULL) Error(nmfc);

	for (z=0; z<Nz; z++) {
	    ReadImageFile(path[z],&nx,&ny,NULL,&slice,NULL);

	    if (nx==Nx && nx==Ny)
		cell[z]=slice;
	    else {
		if ((cell[z]=(Cell **)MA(Ny,*cell))==NULL) Error(nmfc);

		for (y=0; y<Ny; y++) {
		    if ((cell[z][y]=(Cell *)MA(Nx,**cell))==NULL) Error(nmfc);

		    for (x=0; x<Nx; x++)
			cell[z][y][x]=(x<nx && y<ny)?slice[y][x]:0;

		    if (y<ny) free(slice[y]);
		}
		free(slice);
	    }
	}

	if (!Binarization(Nx,Ny,Nz,cell,atoi(argv[3]),
					atoi(argv[4]),&x1,&y1,&z1,&x2,&y2,&z2))
	    Error("no object pixel.");

	if ((last=(Cell **)MA(y2-y1+1,last))==NULL) Error(nmfl);

	last-=y1;
	for (y=y1; y<=y2; y++) {
	    if ((last[y]=(Cell *)MA(x2-x1+1,*last))==NULL) Error(nmfl);

	    last[y]-=x1;
	}

	for (layer=1; layer<CORE; layer++)
	    if (!Erosion(&x1,&y1,&z1,&x2,&y2,&z2,cell,last,layer)) break;

	printf("%d\n",layer);

	if (argc==6 || argc==7) {
	    for (len=0; argv[1][len]!='\0'; len++) ;

	    bps=(layer<2)?1:(layer<4)?2:(layer<16)?4:(layer<256)?8:16;

	    if (cms && cms<1<<bps) {
		fputs("too few colormap (warning).\n",stderr);

		cms=0;
	    }
	    for (z=0; z<Nz; z++) {
		sprintf(name,"%s%s",argv[argc-1],path[z]+len);
		if (cms)
		    StoreImageFile_CM(name,Nx,Ny,bps,R,G,B,cell[z],NULL);
		else
		    StoreImageFile(name,Nx,Ny,bps,cell[z],NULL);
	    }
	}
	return 0;
}
